import asyncio
import time


class ExecutionTimer:
    """
    Context manager for time limiting of accessing to context block.
    Simply sleep `period` secs before next accessing, not analog of Throttler.
    Also it can align to start of minutes.

    Example usage:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_execution_timer.py
    """
    __slots__ = ('_period', '_align_sleep', '_start_time', '_next_time',)

    def __init__(self, period: float = 60., align_sleep: bool = False):
        self._period = period
        self._align_sleep = align_sleep

        self._start_time = 0.
        self._next_time = 0.

    def _start(self):
        curr_time = time.time()
        diff = self._next_time - curr_time
        return diff

    def _exit(self):
        next_time = self._start_time + self._period
        if self._align_sleep:
            next_time -= self._start_time % self._period
        self._next_time = next_time

    def __enter__(self):
        diff = self._start()
        if diff > 0.:
            time.sleep(diff)
        self._start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exit()

    async def __aenter__(self):
        diff = self._start()
        if diff > 0.:
            await asyncio.sleep(diff)
        self._start_time = time.time()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._exit()



import asyncio


class ThrottlerSimultaneous:
    """
    Context manager for limiting simultaneous count of accessing to context block.

    Should be created inside of async loop.

    Example usages:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers.py
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers_aiohttp.py
    """

    __slots__ = ('_semaphore',)

    def __init__(self, count: int):
        self._semaphore = asyncio.Semaphore(count)

    async def __aenter__(self):
        await self._semaphore.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._semaphore.release()



import asyncio
import time
from collections import deque


class Throttler:
    """
    Context manager for limiting rate of accessing to context block.

    Example usages:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers.py
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers_aiohttp.py
    """
    __slots__ = ('_rate_limit', '_period', '_times',)

    def __init__(self, rate_limit: int, period: float = 1.0):
        self._rate_limit = rate_limit
        self._period = period

        self._times = deque()

    async def __aenter__(self):
        while True:
            curr_ts = time.monotonic()
            if len(self._times) < self._rate_limit:
                break

            diff = curr_ts - (self._times[-1] + self._period)
            if diff > 0.:
                self._times.pop()
                break
            await asyncio.sleep(diff)

        self._times.appendleft(curr_ts)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



"""Zero-dependency Python package for simple throttling with asyncio support"""
from .execution_timer import ExecutionTimer
from .throttler import Throttler
from .throttler_simultaneous import ThrottlerSimultaneous
from .timer import Timer

from .decorators import execution_timer, execution_timer_async, throttle, throttle_simultaneous, timer, timer_async

__author__ = 'uburuntu'
__email__ = 'github@rmbk.me'

__license__ = 'MIT'
__version__ = '1.1.3'



from datetime import datetime
from typing import Callable


class Timer:
    """
    Context manager for pretty printing start, end, elapsed and average times.

    Example usage:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_timer.py
    """

    def __init__(self, name: str = None, verbose: bool = False, print_func: Callable = None):
        self.iteration = 1
        self.start_dt = None
        self.elapsed_all = 0.

        self.name = name
        self.verbose = verbose
        self.print = print_func or print

    def __enter__(self):
        self.start_dt = datetime.now()
        if self.verbose:
            self.print(f'{f"#{self.iteration}":>5} | {self.name or "Timer"} | begin: {self.start_dt}')

    def __exit__(self, exc_type, exc_val, exc_tb):
        curr_dt = datetime.now()
        elapsed = (curr_dt - self.start_dt).total_seconds()

        self.elapsed_all += elapsed
        average = self.elapsed_all / self.iteration

        if self.verbose:
            self.print(f'{f"#{self.iteration}":>5} | {self.name or "Timer"} |   end: {curr_dt}, elapsed: {elapsed:.2f} sec, '
                       f'average: {average:.2f} sec\n')
        else:
            self.print(f'{self.name or "Timer"} | elapsed: {elapsed:.2f} sec')

        self.iteration += 1



from functools import wraps
from typing import Callable

from throttler import ExecutionTimer, Throttler, ThrottlerSimultaneous, Timer


def throttle(rate_limit: int, period: float = 1.0):
    def decorator(func):
        throttler = Throttler(rate_limit, period)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            async with throttler:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


def throttle_simultaneous(count: int):
    def decorator(func):
        throttler = ThrottlerSimultaneous(count)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            async with throttler:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


def execution_timer(period: float = 60., align_sleep: bool = False):
    def decorator(func):
        et = ExecutionTimer(period, align_sleep)

        @wraps(func)
        def wrapper(*args, **kwargs):
            with et:
                return func(*args, **kwargs)

        return wrapper

    return decorator


def execution_timer_async(period: float = 60., align_sleep: bool = False):
    def decorator(func):
        et = ExecutionTimer(period, align_sleep)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            async with et:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


def timer(name: str = None, verbose: bool = False, print_func: Callable = None):
    def decorator(func):
        t = Timer(name, verbose, print_func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            with t:
                return func(*args, **kwargs)

        return wrapper

    return decorator


def timer_async(name: str = None, verbose: bool = False, print_func: Callable = None):
    def decorator(func):
        t = Timer(name, verbose, print_func)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            with t:
                return await func(*args, **kwargs)

        return wrapper

    return decorator
